from functools import partial

from ..i18n import fast_translate
from ..tales import TalesEngine
from ..tales import PythonExpr
from ..tales import StringExpr
from ..tales import NotExpr
from ..tales import ExistsExpr
from ..tales import ImportExpr
from ..tal import RepeatDict

from ..template import BaseTemplate
from ..template import BaseTemplateFile

from .program import MacroProgram as Program

try:
    str = unicode
except NameError:
    pass


class PageTemplate(BaseTemplate):
    """Template class for the Chameleon Page Templates language.

    This constructor takes a string input. To read a template source
    from a file on disk, use the file-based constructor.

    To provide a target language for translation, pass
    ``target_language`` as a keyword argument to the render method.

    The default translation function is given as a class attribute
    ``translate``. This may be overriden either on construction, or at
    render-time (in both cases, pass keyword argument ``translate``
    with a valid translation function).

    Content for substitution may be given as encoded input. To enable
    this feature, the ``encoding`` keyword argument can be passed to
    the constructor or render method. The default encoding is the
    system default --- usually ``ascii``.

    Output is always type ``unicode`` (or simply ``str`` on Python 3).
    """

    expression_types = {
        'python': PythonExpr,
        'string': StringExpr,
        'not': NotExpr,
        'exists': ExistsExpr,
        'import': ImportExpr,
        }

    default_expression = 'python'

    translate = staticmethod(fast_translate)

    encoding = None

    mode = "xml"

    def __init__(self, *args, **kwargs):
        super(PageTemplate, self).__init__(*args, **kwargs)
        self.__dict__.update(kwargs)
        self.macros = Macros(self)

    @property
    def engine(self):
        return TalesEngine(self.expression_types, self.default_expression)

    def parse(self, body):
        escape = True if self.mode == "xml" else False
        return Program(body, self.mode, self.filename, escape=escape)

    def render(self, encoding=None, translate=None, target_language=None, **k):
        translate = translate if translate is not None else self.translate

        # Curry language parameter if non-trivial
        if target_language is not None:
            translate = partial(translate, target_language=target_language)

        encoding = encoding if encoding is not None else self.encoding
        if encoding is not None:
            txl = translate

            def translate(msgid, **kwargs):
                if isinstance(msgid, bytes):
                    msgid = str(msgid, encoding)
                return txl(msgid, **kwargs)

            def decode(inst):
                return str(inst, encoding)
        else:
            decode = str

        setdefault = k.setdefault
        setdefault("template", self)
        setdefault("macros", self.macros)
        setdefault("repeat", RepeatDict())
        setdefault("translate", translate)
        setdefault("convert", translate)
        setdefault("decode", decode)
        setdefault("nothing", None)

        return super(PageTemplate, self).render(**k)

    def include(self, stream, econtext, rcontext):
        self.cook_check()
        self._render(stream, econtext, rcontext)


class PageTemplateFile(PageTemplate, BaseTemplateFile):
    pass


class PageTextTemplate(PageTemplate):
    mode = "text"


class PageTextTemplateFile(PageTemplateFile):
    mode = "text"

    def render(self, *args, **kwargs):
        result = super(PageTextTemplateFile, self).render(*args, **kwargs)
        return result.encode(self.encoding or 'utf-8')


class Macro(object):
    __slots__ = "include",

    def __init__(self, render):
        self.include = render


class Macros(object):
    __slots__ = "template",

    def __init__(self, template):
        self.template = template

    def __getitem__(self, name):
        name = name.replace('-', '_')
        self.template.cook_check()

        try:
            function = getattr(self.template, "_render_%s" % name)
        except AttributeError:
            raise KeyError(
                "Macro does not exist: '%s'." % name)

        return Macro(function)

    @property
    def names(self):
        self.template.cook_check()

        result = []
        for name in self.template.__dict__:
            if name.startswith('_render_'):
                result.append(name[8:])
        return result



import re
import sys

try:
    import ast
except ImportError:
    from chameleon import ast24 as ast

from .astutil import parse
from .astutil import store
from .astutil import load
from .astutil import ItemLookupOnAttributeErrorVisitor
from .codegen import TemplateCodeGenerator
from .codegen import template
from .codegen import reverse_builtin_map
from .astutil import Builtin
from .astutil import Symbol
from .exc import ExpressionError
from .utils import decode_htmlentities
from .utils import resolve_dotted
from .tokenize import Token
from .parser import groupdict

try:
    from .py26 import lookup_attr
except SyntaxError:
    from .py25 import lookup_attr


split_parts = re.compile(r'(?<!\\)\|')
match_prefix = re.compile(r'^\s*([a-z\-_]+):').match


try:
    from __builtin__ import basestring
except ImportError:
    basestring = str


def resolve_global(value):
    name = reverse_builtin_map.get(value)
    if name is not None:
        return Builtin(name)

    return Symbol(value)


def dummy_engine(string, target):
    pos = string.find('{')
    if pos != -1:
        raise ExpressionError("Bad input", string)
    return [ast.Assign(targets=[target], value=ast.Str(s=string))]


def test(expression, engine=dummy_engine):
    body = expression(store("result"), engine)
    module = ast.Module(body)
    module = ast.fix_missing_locations(module)
    env = {
        'rcontext': {},
        }
    source = TemplateCodeGenerator(module).code
    code = compile(source, '<string>', 'exec')
    exec(code, env)
    result = env["result"]

    if isinstance(result, basestring):
        result = str(result)

    return result


def transform_attribute(node):
    return template(
        "lookup(object, name)",
        lookup=Symbol(lookup_attr),
        object=node.value,
        name=ast.Str(s=node.attr),
        mode="eval"
        )


class TalesExpr(object):
    """Base class.

    This class helps implementations for the Template Attribute
    Language Expression Syntax (TALES).

    The syntax evaluates one or more expressions, separated by '|'
    (pipe). The first expression that succeeds, is returned.

    Expression:

      expression    := (type ':')? line ('|' expression)?
      line          := .*

    Expression lines may not contain the pipe character unless
    escaped. It has a special meaning:

    If the expression to the left of the pipe fails (raises one of the
    exceptions listed in ``catch_exceptions``), evaluation proceeds to
    the expression(s) on the right.

    Subclasses must implement ``translate`` which assigns a value for
    a given expression.

    >>> class PythonPipeExpr(TalesExpr):
    ...     def translate(self, expression, target):
    ...         compiler = PythonExpr(expression)
    ...         return compiler(target, dummy_engine)

    >>> test(PythonPipeExpr('foo | bar | 42'))
    42

    >>> test(PythonPipeExpr('foo|42'))
    42

    """

    exceptions = NameError, \
                 ValueError, \
                 AttributeError, \
                 LookupError, \
                 TypeError

    def __init__(self, expression):
        self.expression = expression

    def __call__(self, target, engine):
        remaining = self.expression
        assignments = []

        while remaining:
            if match_prefix(remaining) is not None:
                assignment = engine(remaining, target)
                remaining = ""
            else:
                for m in split_parts.finditer(remaining):
                    expression = remaining[:m.start()]
                    remaining = remaining[m.end():]
                    break
                else:
                    expression = remaining
                    remaining = ""

                expression = expression.replace('\\|', '|')
                assignment = self.translate(expression, target)
            assignments.append(assignment)

        if not assignments:
            assignments.append(
                self.translate(remaining, target)
                )

        for i, assignment in enumerate(reversed(assignments)):
            if i == 0:
                body = assignment
            else:
                body = [ast.TryExcept(
                    body=assignment,
                    handlers=[ast.ExceptHandler(
                        type=ast.Tuple(
                            elts=map(resolve_global, self.exceptions),
                            ctx=ast.Load()),
                        name=None,
                        body=body,
                        )],
                    )]

        return body

    def translate(self, expression, target):
        """Return statements that assign a value to ``target``."""

        raise NotImplementedError(
            "Must be implemented by a subclass.")


class PathExpr(TalesExpr):
    """Path expression compiler.

    Syntax::

        PathExpr ::= Path [ '|' Path ]*
        Path ::= variable [ '/' URL_Segment ]*
        variable ::= Name

    For example::

        request/cookies/oatmeal
        nothing
        here/some-file 2001_02.html.tar.gz/foo
        root/to/branch | default

    When a path expression is evaluated, it attempts to traverse
    each path, from left to right, until it succeeds or runs out of
    paths. To traverse a path, it first fetches the object stored in
    the variable. For each path segment, it traverses from the current
    object to the subobject named by the path segment.

    Once a path has been successfully traversed, the resulting object
    is the value of the expression. If it is a callable object, such
    as a method or class, it is called.

    The semantics of traversal (and what it means to be callable) are
    implementation-dependent (see the ``translate`` method).
    """

    def translate(self, expression, target):
        raise NotImplementedError(
            "Path expressions are not yet implemented. "
            "It's unclear whether a general implementation "
            "can be devised.")


class PythonExpr(TalesExpr):
    """Python expression compiler.

    >>> test(PythonExpr('2 + 2'))
    4

    The Python expression is a TALES expression. That means we can use
    the pipe operator:

    >>> test(PythonExpr('foo | 2 + 2 | 5'))
    4

    To include a pipe character, use a backslash escape sequence:

    >>> test(PythonExpr('\"\|\"'))
    '|'

    """

    transform = ItemLookupOnAttributeErrorVisitor(transform_attribute)

    def translate(self, expression, target):
        string = expression.strip().replace('\n', ' ')
        decoded = decode_htmlentities(string)

        try:
            value = parse(decoded, 'eval').body
        except SyntaxError:
            exc = sys.exc_info()[1]
            raise ExpressionError(exc.msg, decoded)

        # Transform attribute lookups to allow fallback to item lookup
        self.transform.visit(value)

        return [ast.Assign(targets=[target], value=value)]


class ImportExpr(object):
    re_dotted = re.compile(r'^[A-Za-z.]+$')

    def __init__(self, expression):
        self.expression = expression

    def __call__(self, target, engine):
        string = self.expression.strip().replace('\n', ' ')
        value = template(
            "RESOLVE(NAME)",
            RESOLVE=Symbol(resolve_dotted),
            NAME=ast.Str(s=string),
            mode="eval",
            )
        return [ast.Assign(targets=[target], value=value)]


class NotExpr(object):
    """Negates the expression.

    >>> def engine(expression, target):
    ...     expr = PythonExpr(expression)
    ...     return expr(target, None)

    >>> test(NotExpr('False'), engine)
    True
    >>> test(NotExpr('True'), engine)
    False
    """

    def __init__(self, expression):
        self.expression = expression

    def __call__(self, target, engine):
        body = engine(self.expression, target)
        return body + template("target = not target", target=target)


class IdentityExpr(object):
    """Identity expression.

    Exists to demonstrate the interface.

    >>> test(IdentityExpr('Hello world!'))
    'Hello world!'
    """

    def __init__(self, expression):
        self.expression = expression

    def __call__(self, target, engine):
        return engine(self.expression, target)


class StringExpr(object):
    """Similar to the built-in ``string.Template``, but uses an
    expression engine to support pluggable string substitution
    expressions.

    Expr string:

      string       := (text | substitution) (string)?
      substitution := ('$' variable | '${' expression '}')
      text         := .*

    In other words, an expression string can contain multiple
    substitutions. The text- and substitution parts will be
    concatenated back into a string.

    >>> test(StringExpr('Hello ${world}!'))
    'Hello world!'

    We can escape interpolation using the standard escaping
    syntax:

    >>> test(StringExpr('\\${Hello}'))
    '\\\${Hello}'

    Multiple interpolations in one:

    >>> test(StringExpr('Hello ${a}${b}${c}!'))
    'Hello abc!'

    Here's a more involved example taken from a javascript source:

    >>> result = test(StringExpr(\"\"\"
    ... function(oid) {
    ...     $('#' + oid).autocomplete({source: ${source}});
    ... }
    ... \"\"\"))

    >>> 'source: source' in result
    True

    In the above examples, the expression is evaluated using the
    dummy engine which just returns the input as a string.

    As an example, we'll implement an expression engine which
    instead counts the number of characters in the expresion and
    returns an integer result.

    >>> def engine(expression, target):
    ...     return [ast.Assign(targets=[target],
    ...             value=ast.Num(n=len(expression)))]

    This will demonstrate how the string expression coerces the
    input to a string.

    >>> expr = StringExpr(
    ...    'There are ${hello world} characters in \"hello world\"')

    We evaluate the expression using the new engine:

    >>> test(expr, engine)
    'There are 11 characters in \"hello world\"'
    """

    regex = re.compile(
        r'(?<!\\)\$({(?P<expression>.*)}|(?P<variable>[A-Za-z][A-Za-z0-9_]*))')

    def __init__(self, expression):
        # The code relies on the expression being a token string
        if not isinstance(expression, Token):
            expression = Token(expression, 0)

        self.expression = expression

    def __call__(self, name, engine):
        """The strategy is to find possible expression strings and
        call the ``validate`` function of the parser to validate.

        For every possible starting point, the longest possible
        expression is tried first, then the second longest and so
        forth.

        Example 1:

          ${'expressions use the ${<expression>} format'}

        The entire expression is attempted first and it is also the
        only one that validates.

        Example 2:

          ${'Hello'} ${'world!'}

        Validation of the longest possible expression (the entire
        string) will fail, while the second round of attempts,
        ``${'Hello'}`` and ``${'world!'}`` respectively, validate.

        """

        body = []
        nodes = []
        text = self.expression

        while text:
            matched = text
            m = self.regex.search(matched)
            if m is None:
                nodes.append(ast.Str(s=text))
                break

            part = text[:m.start()]
            text = text[m.start():]

            if part:
                node = ast.Str(s=part)
                nodes.append(node)

            if not body:
                target = name
            else:
                target = store("%s_%d" % (name.id, text.pos))

            while True:
                d = groupdict(m, matched)
                string = d["expression"] or d["variable"] or ""

                try:
                    body += engine(string, target)
                except ExpressionError:
                    matched = matched[m.start():m.end() - 1]
                    m = self.regex.search(matched)
                    if m is None:
                        raise
                else:
                    break

            # if this is the first expression, use the provided
            # assignment name; otherwise, generate one (here based
            # on the string position)
            node = load(target.id)
            nodes.append(node)
            text = text[len(m.group()):]

        if len(nodes) == 1:
            target = nodes[0]
        else:
            nodes = [
                template("NODE if NODE is not None else ''", NODE=node, mode="eval")
                for node in nodes
                ]

            target = ast.BinOp(
                left=ast.Str(s="%s" * len(nodes)),
                op=ast.Mod(),
                right=ast.Tuple(elts=nodes, ctx=ast.Load()))

        body += [ast.Assign(targets=[name], value=target)]
        return body


class ExistsExpr(object):
    """Boolean wrapper.

    Return 0 if the expression results in an exception, otherwise 1.

    As a means to generate exceptions, we set up an expression engine
    which evaluates the provided expression using Python:

    >>> def engine(expression, target):
    ...     expr = PythonExpr(expression)
    ...     return expr(target, None)

    >>> test(ExistsExpr('int(0)'), engine)
    1
    >>> test(ExistsExpr('int(None)'), engine)
    0

    """

    exceptions = AttributeError, LookupError, TypeError, NameError, KeyError

    def __init__(self, expression):
        self.expression = expression

    def __call__(self, target, engine):
        ignore = store("_ignore")
        body = engine(self.expression, ignore)

        classes = map(resolve_global, self.exceptions)

        return [
            ast.TryExcept(
                body=body,
                handlers=[ast.ExceptHandler(
                    type=ast.Tuple(elts=classes, ctx=ast.Load()),
                    name=None,
                    body=template("target = 0", target=target),
                    )],
                orelse=template("target = 1", target=target)
                )
            ]


class TalesEngine(object):
    """Expression engine.

    This test demonstrates how to configure and invoke the engine.

    >>> engine = TalesEngine({
    ...     'python': PythonExpr,
    ...     'not': NotExpr,
    ...     'exists': ExistsExpr,
    ...     'string': StringExpr,
    ...     }, 'python')

    An expression evaluation function:

    >>> eval = lambda expression: test(IdentityExpr(expression), engine)

    We have provided 'python' as the default expression type. This
    means that when no prefix is given, the expression is evaluated as
    a Python expression:

    >>> eval('not False')
    True

    Note that the ``type`` prefixes bind left. If ``not`` and
    ``exits`` are two expression type prefixes, consider the
    following::

    >>> eval('not: exists: int(None)')
    True

    The pipe operator binds right. In the following example, but
    arguments are evaluated against ``not: exists: ``.

    >>> eval('not: exists: help')
    False

    >>> eval('string:test ${1}${2}')
    'test 12'

    This expression would check whether one of the two objects exist,
    then negate the result.
    """

    def __init__(self, factories, default):
        self.factories = factories
        self.default = default

    def __call__(self, expression, target):
        m = match_prefix(expression)
        if m is not None:
            prefix = m.group(1)
            expression = expression[m.end():]
        else:
            prefix = self.default

        factory = self.get_factory(prefix)
        compiler = factory(expression)
        return compiler(target, self)

    def get_factory(self, prefix=None):
        if prefix is None:
            prefix = self.default

        try:
            return self.factories[prefix]
        except KeyError:
            exc = sys.exc_info()[1]
            raise LookupError(
                "Unknown expression type: %s." % str(exc)
                )



from .astutil import Node


class UseExternalMacro(Node):
    """Extend external macro."""

    _fields = "expression", "slots", "extend"


class Sequence(Node):
    """Element sequence."""

    _fields = "items",


class Content(Node):
    """Content substitution.

    If ``msgid`` is non-trivial, the content will be subject to
    translation. The provided ``default`` is the default content.
    """

    _fields = "expression", "msgid", "escape"


class Expression(Node):
    """String expression for evaluation by expression engine."""

    _fields = "value",

    def __repr__(self):
        try:
            line, column = self.value.location
        except AttributeError:
            line, column = 0, 0

        return "<Expression %r (%d:%d)>" % (self.value, line, column)


class Negate(Node):
    """Wraps an expression with a negation."""

    _fields = "value",


class Element(Node):
    """XML element."""

    _fields = "start", "end", "content"


class Attribute(Node):
    """Element attribute."""

    _fields = "name", "expression", "quote", "eq", "space", "escape"


class Start(Node):
    """Start-tag."""

    _fields = "name", "prefix", "suffix", "attributes"


class End(Node):
    """End-tag."""

    _fields = "name", "space", "prefix", "suffix"


class Condition(Node):
    """Node visited only if some condition holds."""

    _fields = "expression", "node", "orelse"


class Identity(Node):
    """Condition expression that is true on identity."""

    _fields = "expression", "value"


class Equality(Node):
    """Condition expression that is true on equality."""

    _fields = "expression", "value"


class Marker(Node):
    """Represents a marker object.

    The ``name`` string must be unique across the template program.
    """

    _fields = "name",


class Cache(Node):
    """Cache (evaluate only once) the value of ``expression`` inside
    ``node``.
    """

    _fields = "expressions", "node"


class Assignment(Node):
    """Variable assignment."""

    _fields = "names", "expression", "local"


class Define(Node):
    """Variable definition in scope."""

    _fields = "assignments", "node"


class Repeat(Assignment):
    """Iterate over provided assignment and repeat body."""

    _fields = "names", "expression", "local", "whitespace", "node"


class Macro(Node):
    """Macro definition."""

    _fields = "name", "body"


class Program(Node):
    _fields = "name", "body"


class Text(Node):
    """Static text output."""

    _fields = "value",


class Interpolation(Text):
    """String interpolation output."""

    _fields = "value", "escape"


class Translate(Node):
    """Translate node."""

    _fields = "msgid", "node"


class Name(Node):
    """Translation name."""

    _fields = "name", "node"


class Domain(Node):
    """Update translation domain."""

    _fields = "name", "node"


class OnError(Node):
    _fields = "fallback", "node"


class UseInternalMacro(Node):
    """Use internal macro (defined inside same program)."""

    _fields = "name",


class FillSlot(Node):
    """Fill a macro slot."""

    _fields = "name", "node"


class DefineSlot(Node):
    """Define a macro slot."""

    _fields = "name", "node"



import re
import sys
import itertools
import logging
import threading

try:
    import ast
except ImportError:
    from chameleon import ast24 as ast

try:
    fast_string = str
    str = unicode
    bytes = fast_string
except NameError:
    long = int
    basestring = str
    fast_string = str

try:
    import __builtin__ as builtins
except ImportError:
    import builtins

from .astutil import load
from .astutil import store
from .astutil import param
from .astutil import swap
from .astutil import subscript
from .astutil import node_annotations
from .astutil import NameLookupRewriteVisitor
from .astutil import Comment
from .astutil import Symbol
from .astutil import Builtin

from .codegen import TemplateCodeGenerator
from .codegen import template

from .tales import StringExpr
from .i18n import fast_translate

from .nodes import Text
from .nodes import Expression

from .config import DEBUG_MODE
from .exc import TranslationError
from .utils import DebuggingOutputStream
from .utils import Placeholder
from .tokenize import Token


log = logging.getLogger('chameleon.compiler')

COMPILER_INTERNALS_OR_DISALLOWED = set([
    "stream",
    "append",
    "econtext",
    "rcontext",
    "translate",
    "decode",
    "convert",
    "str",
    "int",
    "float",
    "long",
    "len",
    "None",
    "True",
    "False",
    ])


RE_MANGLE = re.compile('[\-: ]')

if DEBUG_MODE:
    LIST = template("cls()", cls=DebuggingOutputStream, mode="eval")
else:
    LIST = template("[]", mode="eval")


def identifier(prefix, suffix=None):
    return "_%s_%s" % (prefix, mangle(suffix or id(prefix)))


def mangle(string):
    return RE_MANGLE.sub('_', str(string)).replace('\n', '')


def load_econtext(name):
    return template("getitem(KEY)", KEY=ast.Str(s=name), mode="eval")


def store_econtext(name):
    name = fast_string(name)
    return subscript(name, load("econtext"), ast.Store())


def store_rcontext(name):
    name = fast_string(name)
    return subscript(name, load("rcontext"), ast.Store())


@template
def emit_node(node):  # pragma: no cover
    append(node)


@template
def emit_node_if_non_trivial(node):  # pragma: no cover
    if node is not None:
        append(node)


@template
def emit_convert(target, native=bytes, str=str, long=long):  # pragma: no cover
    if target is not None:
        _tt = type(target)

        if _tt is int or _tt is float or _tt is long:
            target = native(target)
        elif _tt is native:
            target = decode(target)
        elif _tt is not str:
            try:
                target = target.__html__
            except AttributeError:
                target = convert(target)
            else:
                target = target()


@template
def emit_translate(target, msgid, default=None):  # pragma: no cover
    target = translate(msgid, default=default, domain=_i18n_domain)


@template
def emit_convert_and_escape(
    target, msgid, quote=ast.Str(s="\0"), str=str, long=long,
    native=bytes):  # pragma: no cover
    if target is None:
        pass
    elif target is False:
        target = None
    else:
        _tt = type(target)

        if _tt is int or _tt is float or _tt is long:
            target = str(target)
        else:
            try:
                if _tt is native:
                    target = decode(msgid)
                elif _tt is not str:
                    try:
                        target = target.__html__
                    except:
                        target = convert(msgid)
                    else:
                        raise RuntimeError
            except RuntimeError:
                target = target()
            else:
                if target is not None and re_needs_escape(target) is not None:
                    # Character escape
                    if '&' in target:
                        # If there's a semicolon in the string, then
                        # it might be part of an HTML entity. We
                        # replace the ampersand character with its
                        # HTML entity counterpart only if it's
                        # precedes an HTML entity string.
                        if ';' in target:
                            target = re_amp.sub('&amp;', target)

                        # Otherwise, it's safe to replace all
                        # ampersands:
                        else:
                            target = target.replace('&', '&amp;')

                    if '<' in target:
                        target = target.replace('<', '&lt;')
                    if '>' in target:
                        target = target.replace('>', '&gt;')
                    if quote in target:
                        target = target.replace(quote, '&#34;')


class ExpressionCompiler(object):
    """Internal wrapper around a TALES-engine.

    In addition to TALES expressions (strings which may appear wrapped
    in an ``Expression`` node), this compiler also supports other
    expression node types.

    Used internally be the compiler.
    """

    initial = COMPILER_INTERNALS_OR_DISALLOWED

    global_fallback = set(builtins.__dict__)

    def __init__(self, engine, cache, markers=None):
        self.engine = engine
        self.cache = cache
        self.markers = markers

    def __call__(self, expression, target):
        if isinstance(target, basestring):
            target = store(target)

        stmts = self.translate(expression, target)

        # Apply dynamic name rewrite transform to each statement
        visitor = NameLookupRewriteVisitor(self._dynamic_transform)

        for stmt in stmts:
            visitor(stmt)

        return stmts

    def translate(self, expression, target):
        if isinstance(target, basestring):
            target = store(target)

        cached = self.cache.get(expression)

        if cached is not None:
            stmts = [ast.Assign(targets=[target], value=cached)]
        elif isinstance(expression, ast.expr):
            stmts = [ast.Assign(targets=[target], value=expression)]
            self.cache[expression] = target
        else:
            # The engine interface supports simple strings, which
            # default to expression nodes
            if isinstance(expression, basestring):
                expression = Expression(expression)

            kind = type(expression).__name__
            visitor = getattr(self, "visit_%s" % kind)
            stmts = visitor(expression, target)

            # Add comment
            target_id = getattr(target, "id", target)
            comment = Comment(" %r -> %s" % (expression, target_id))
            stmts.insert(0, comment)

        return stmts

    @classmethod
    def _dynamic_transform(cls, node):
        # Don't rewrite nodes that have an annotation
        annotation = node_annotations.get(node)
        if annotation is not None:
            return node

        name = node.id

        # Don't rewrite names that begin with an underscore; they are
        # internal and can be assumed to be locally defined. This
        # policy really should be part of the template program, not
        # defined here in the compiler.
        if name.startswith('_') or name in cls.initial:
            return node

        if isinstance(node.ctx, ast.Store):
            return store_econtext(name)

        # If the name is a Python global, first try acquiring it from
        # the dynamic context, then fall back to the global.
        if name in cls.global_fallback:
            return template(
                "get(key, name)",
                mode="eval",
                key=ast.Str(s=name),
                name=name
                )

        # Otherwise, simply acquire it from the dynamic context.
        return load_econtext(name)

    def visit_Expression(self, node, target):
        stmts = self.engine(node.value, target)

        try:
            line, column = node.value.location
            filename = node.value.filename
        except AttributeError:
            line, column = 0, 0
            filename = "<string>"

        return [ast.TryExcept(
            body=stmts,
            handlers=[ast.ExceptHandler(
                body=template(
                    "rcontext.setdefault('__error__', [])."
                    "append((expression, line, col, src, sys.exc_info()[1]))\n"
                    "raise",
                    expression=ast.Str(s=node.value),
                    line=ast.Num(n=line),
                    col=ast.Num(n=column),
                    src=ast.Str(s=filename),
                    sys=Symbol(sys),
                    ),
                )],
            )]

    def visit_Negate(self, node, target):
        return self.translate(node.value, target) + \
               template("TARGET = not TARGET", TARGET=target)

    def visit_Marker(self, node, target):
        self.markers.add(node.name)

        return [ast.Assign(targets=[target],
                           value=load("_marker_%s" % node.name))]

    def visit_Identity(self, node, target):
        expression = self.translate(node.expression, "_expression")
        value = self.translate(node.value, "_value")

        return expression + value + \
               template("TARGET = _expression is _value", TARGET=target)

    def visit_Equality(self, node, target):
        expression = self.translate(node.expression, "_expression")
        value = self.translate(node.value, "_value")

        return expression + value + \
               template("TARGET = _expression == _value", TARGET=target)

    def visit_Interpolation(self, node, target):
        def engine(expression, target):
            node = Expression(expression)
            return self.translate(node, target)

        expression = StringExpr(node.value)
        return expression(target, engine)

    def visit_Translate(self, node, target):
        if node.msgid is not None:
            msgid = ast.Str(s=node.msgid)
        else:
            msgid = target
        return self.translate(node.node, target) + \
               emit_translate(target, msgid, default=target)

    def visit_Static(self, node, target):
        value = load("dummy")
        node_annotations[value] = node
        return [ast.Assign(targets=[target], value=value)]

    def visit_Builtin(self, node, target):
        value = load("dummy")
        node_annotations[value] = node
        return [ast.Assign(targets=[target], value=value)]


class Compiler(object):
    """Generic compiler class.

    Iterates through nodes and yields Python statements which form a
    template program.
    """

    exceptions = NameError, \
                 ValueError, \
                 AttributeError, \
                 LookupError, \
                 TypeError

    defaults = {
        'translate': Symbol(fast_translate),
        'decode': Builtin("str"),
        'convert': Builtin("str"),
        }

    lock = threading.Lock()

    def __init__(self, engine, node):
        self._scopes = [set()]
        self._expression_cache = {}
        self._translations = []
        self._markers = set()

        self._engine = ExpressionCompiler(
            engine,
            self._expression_cache,
            self._markers
            )

        if isinstance(node_annotations, dict):
            self.lock.acquire()
            backup = node_annotations.copy()
        else:
            backup = None

        try:
            module = ast.Module([])
            module.body += self.visit(node)
            ast.fix_missing_locations(module)
            generator = TemplateCodeGenerator(module)
        finally:
            if backup is not None:
                node_annotations.clear()
                node_annotations.update(backup)
                self.lock.release()

        self.code = generator.code

    def visit(self, node):
        if node is None:
            return ()
        kind = type(node).__name__
        visitor = getattr(self, "visit_%s" % kind)
        iterator = visitor(node)
        return list(iterator)

    def visit_Sequence(self, node):
        for item in node.items:
            for stmt in self.visit(item):
                yield stmt

    def visit_Element(self, node):
        for stmt in self.visit(node.start):
            yield stmt

        for stmt in self.visit(node.content):
            yield stmt

        if node.end is not None:
            for stmt in self.visit(node.end):
                yield stmt

    def visit_Program(self, node):
        return self.visit_MacroProgram(node)

    def visit_MacroProgram(self, node):
        body = []

        body += template("import re")
        body += template("import functools")
        body += template("_marker = object()")
        body += template(
            r"g_re_amp = re.compile(r'&(?!([A-Za-z]+|#[0-9]+);)')"
        )
        body += template(
            r"g_re_needs_escape = re.compile(r'[&<>\"\']').search")

        body += template(
            r"re_whitespace = "
            r"functools.partial(re.compile('\s+').sub, ' ')",
        )

        # Visit defined macros
        for macro in getattr(node, "macros", ()):
            body += self.visit(macro)

        # Visit (implicit) main macro
        body += self.visit_Macro(node)

        # Prepend module-wide marker values
        for marker in self._markers:
            body[:] = template(
                "MARKER = CLS()",
                MARKER=store("_marker_%s" % marker),
                CLS=Placeholder,
                ) + body

        return body

    def visit_Macro(self, node):
        body = []

        # Initialization
        body += template("append = stream.append")
        body += template("getitem = econtext.__getitem__")
        body += template("get = econtext.get")
        body += template("_i18n_domain = None")
        body += template("re_amp = g_re_amp")
        body += template("re_needs_escape = g_re_needs_escape")

        # Resolve defaults
        for name in self.defaults:
            body += template(
                "NAME = getitem(KEY)",
                NAME=name, KEY=ast.Str(s=name)
            )

        # Visit macro body
        body += itertools.chain(*tuple(map(self.visit, node.body)))

        function_name = "render" if node.name is None else \
                        "render_%s" % mangle(node.name)

        function = ast.FunctionDef(
            name=function_name, args=ast.arguments(
                args=[
                    param("stream"),
                    param("econtext"),
                    param("rcontext"),
                    ],
                defaults=(),
            ),
            body=body
            )

        yield function

    def visit_Text(self, node):
        return emit_node(ast.Str(s=node.value))

    def visit_Domain(self, node):
        backup = "_previous_i18n_domain_%d" % id(node)
        return template("BACKUP = _i18n_domain", BACKUP=backup) + \
               template("_i18n_domain = NAME", NAME=ast.Str(s=node.name)) + \
               self.visit(node.node) + \
               template("_i18n_domain = BACKUP", BACKUP=backup)

    def visit_OnError(self, node):
        body = []

        fallback = identifier("_fallback")
        body += template("fallback = len(stream)", fallback=fallback)

        body += [ast.TryExcept(
            body=self.visit(node.node),
            handlers=[ast.ExceptHandler(
                type=ast.Tuple(
                    elts=[Builtin(cls.__name__) for cls in self.exceptions],
                    ctx=ast.Load()),
                name=None,
                body=(template("del stream[fallback:]", fallback=fallback) + \
                      self.visit(node.fallback)
                      ),
                )]
            )]

        return body

    def visit_Content(self, node):
        name = "_content"
        body = self._engine(node.expression, store(name))

        # content conversion steps
        if node.msgid is not None:
            output = emit_translate(name, name)
        elif node.escape:
            output = emit_convert_and_escape(name, name)
        else:
            output = emit_convert(name)

        body += output
        body += template("if NAME is not None: append(NAME)", NAME=name)

        return body

    def visit_Interpolation(self, node):
        if node.escape:
            def convert(target):
                return emit_convert_and_escape(target, target)
        else:
            convert = emit_convert

        def engine(expression, target):
            node = Expression(expression)
            return self._engine(node, target) + \
                   convert(target)

        expression = StringExpr(node.value)

        name = identifier("content")

        return expression(store(name), engine) + \
               emit_node_if_non_trivial(load(name))

    def visit_Assignment(self, node):
        for name in node.names:
            if name in COMPILER_INTERNALS_OR_DISALLOWED:
                raise TranslationError(
                    "Name disallowed by compiler: %s." % name
                    )

        assignment = self._engine(node.expression, store("_value"))

        if len(node.names) != 1:
            target = ast.Tuple(
                elts=[store_econtext(name) for name in node.names],
                ctx=ast.Store(),
            )
        else:
            target = store_econtext(node.names[0])

        assignment.append(ast.Assign(targets=[target], value=load("_value")))

        for name in node.names:
            if not node.local:
                assignment += template(
                    "rcontext[KEY] = _value", KEY=ast.Str(s=fast_string(name))
                    )

        return assignment

    def visit_Define(self, node):
        scope = set(self._scopes[-1])
        self._scopes.append(scope)

        for assignment in node.assignments:
            if assignment.local:
                for stmt in self._enter_assignment(assignment.names):
                    yield stmt

            for stmt in self.visit(assignment):
                yield stmt

        for stmt in self.visit(node.node):
            yield stmt

        for assignment in node.assignments:
            if assignment.local:
                for stmt in self._leave_assignment(assignment.names):
                    yield stmt

        self._scopes.pop()

    def visit_Omit(self, node):
        return self.visit_Condition(node)

    def visit_Condition(self, node):
        target = "_condition"
        assignment = self._engine(node.expression, target)

        assert assignment

        for stmt in assignment:
            yield stmt

        body = self.visit(node.node) or [ast.Pass()]

        orelse = getattr(node, "orelse", None)
        if orelse is not None:
            orelse = self.visit(orelse)

        test = load(target)

        yield ast.If(test=test, body=body, orelse=orelse)

    def visit_Translate(self, node):
        """Translation.

        Visit items and assign output to a default value.

        Finally, compile a translation expression and use either
        result or default.
        """

        body = []

        # Track the blocks of this translation
        self._translations.append(set())

        # Prepare new stream
        append = identifier("append", id(node))
        stream = identifier("stream", id(node))
        body += template("s = new_list", s=stream, new_list=LIST) + \
                template("a = s.append", a=append, s=stream)

        # Visit body to generate the message body
        code = self.visit(node.node)
        swap(ast.Suite(body=code), load(append), "append")
        body += code

        # Reduce white space and assign as message id
        msgid = identifier("msgid", id(node))
        body += template(
            "msgid = re_whitespace(''.join(stream)).strip()",
            msgid=msgid, stream=stream
        )

        default = msgid

        # Compute translation block mapping if applicable
        names = self._translations[-1]
        if names:
            keys = []
            values = []

            for name in names:
                stream, append = self._get_translation_identifiers(name)
                keys.append(ast.Str(s=name))
                values.append(load(stream))

                # Initialize value
                body.insert(
                    0, ast.Assign(
                        targets=[store(stream)],
                        value=ast.Str(s=fast_string(""))))

            mapping = ast.Dict(keys=keys, values=values)
        else:
            mapping = None

        # if this translation node has a name, use it as the message id
        if node.msgid:
            msgid = ast.Str(s=node.msgid)

        # emit the translation expression
        body += template(
            "append(translate("
            "msgid, mapping=mapping, default=default, domain=_i18n_domain))",
            msgid=msgid, default=default, mapping=mapping
            )

        # pop away translation block reference
        self._translations.pop()

        return body

    def visit_Start(self, node):
        try:
            line, column = node.prefix.location
        except AttributeError:
            line, column = 0, 0

        yield Comment(
            " %s%s ... (%d:%d)\n"
            " --------------------------------------------------------" % (
                node.prefix, node.name, line, column))

        if node.attributes:
            for stmt in emit_node(ast.Str(s=node.prefix + node.name)):
                yield stmt

            for attribute in node.attributes:
                for stmt in self.visit(attribute):
                    yield stmt

            for stmt in emit_node(ast.Str(s=node.suffix)):
                yield stmt
        else:
            for stmt in emit_node(
                ast.Str(s=node.prefix + node.name + node.suffix)):
                yield stmt

    def visit_End(self, node):
        for stmt in emit_node(ast.Str(
            s=node.prefix + node.name + node.space + node.suffix)):
            yield stmt

    def visit_Attribute(self, node):
        body = []

        target = identifier("attr", node.name)
        body += self._engine(node.expression, store(target))

        if node.escape:
            body += emit_convert_and_escape(
                target, target, quote=ast.Str(s=node.quote)
                )

        f = node.space + node.name + node.eq + node.quote + "%s" + node.quote
        body += template(
            "if TARGET is not None: append(FORMAT % TARGET)",
            FORMAT=ast.Str(s=f),
            TARGET=target,
            )

        return body

    def visit_Cache(self, node):
        body = []

        for expression in node.expressions:
            name = identifier("cache", id(expression))
            target = store(name)

            # Skip re-evaluation
            if self._expression_cache.get(expression):
                continue

            body += self._engine(expression, target)
            self._expression_cache[expression] = target

        body += self.visit(node.node)

        return body

    def visit_UseInternalMacro(self, node):
        if node.name is None:
            render = "render"
        else:
            render = "render_%s" % mangle(node.name)

        return template("f(stream, econtext.copy(), rcontext)", f=render) + \
               template("econtext.update(rcontext)")

    def visit_DefineSlot(self, node):
        name = "_slot_%s" % mangle(node.name)
        body = self.visit(node.node)

        return [
            ast.TryExcept(
                body=template("_slot = getitem(KEY).pop()",
                              KEY=ast.Str(s=name)),
                handlers=[ast.ExceptHandler(
                    body=body or [ast.Pass()],
                    )],
                orelse=template("_slot(stream, econtext.copy(), econtext)"),
                )
            ]

    def visit_Name(self, node):
        """Translation name."""

        if self._translations is None:
            raise TranslationError(
                "Not allowed outside of translation.", node.name)

        if node.name in self._translations[-1]:
            raise TranslationError(
                "Duplicate translation name: %s." % node.name)

        self._translations[-1].add(node.name)
        body = []

        # prepare new stream
        stream, append = self._get_translation_identifiers(node.name)
        body += template("s = new_list", s=stream, new_list=LIST) + \
                template("a = s.append", a=append, s=stream)

        # generate code
        code = self.visit(node.node)
        swap(ast.Suite(body=code), load(append), "append")
        body += code

        # output msgid
        text = Text('${%s}' % node.name)
        body += self.visit(text)

        # Concatenate stream
        body += template("stream = ''.join(stream)", stream=stream)

        return body

    def visit_UseExternalMacro(self, node):
        callbacks = []

        for slot in node.slots:
            name = "_slot_%s" % mangle(slot.name)

            body = template("getitem = econtext.__getitem__") + \
                   template("get = econtext.get") + \
                   self.visit(slot.node)

            callbacks.append(
                ast.FunctionDef(
                    name=name,
                    args=ast.arguments(
                        args=[
                            param("stream"),
                            param("econtext"),
                            param("rcontext"),
                            param("_i18n_domain"),
                            ],
                        defaults=[load("_i18n_domain")],
                        ),
                    body=body or [ast.Pass()],
                ))

            key = ast.Str(s=name)

            if node.extend:
                update_body = None
            else:
                update_body = template("_slots.append(NAME)", NAME=name)

            callbacks.append(
                ast.TryExcept(
                    body=template("_slots = getitem(KEY)", KEY=key),
                    handlers=[ast.ExceptHandler(
                        body=template(
                            "_slots = econtext[KEY] = [NAME]",
                            KEY=key, NAME=name
                        ))],
                    orelse=update_body
                    ))

        assignment = self._engine(node.expression, store("_macro"))

        return (
            callbacks + \
            assignment + \
            template("_macro.include(stream, econtext.copy(), rcontext)") + \
            template("econtext.update(rcontext)")
            )

    def visit_Repeat(self, node):
        # Used for loop variable definition and restore
        self._scopes.append(set())

        # Variable assignment and repeat key for single- and
        # multi-variable repeat clause
        if node.local:
            contexts = "econtext",
        else:
            contexts = "econtext", "rcontext"

        if len(node.names) > 1:
            targets = [
                ast.Tuple(elts=[
                    subscript(fast_string(name), load(context), ast.Store())
                    for name in node.names], ctx=ast.Store())
                for context in contexts
                ]

            key = ast.Tuple(
                elts=[ast.Str(s=name) for name in node.names],
                ctx=ast.Load())
        else:
            name = node.names[0]
            targets = [
                subscript(fast_string(name), load(context), ast.Store())
                for context in contexts
                ]

            key = ast.Str(s=node.names[0])

        index = identifier("_index", id(node))
        assignment = [ast.Assign(targets=targets, value=load("_item"))]

        # Make repeat assignment in outer loop
        names = node.names
        local = node.local

        outer = self._engine(node.expression, store("_iterator"))

        if local:
            outer[:] = list(self._enter_assignment(names)) + outer

        outer += template(
            "_iterator, INDEX = getitem('repeat')(key, _iterator)",
            key=key, INDEX=index
            )

        # Set a trivial default value for each name assigned to make
        # sure we assign a value even if the iteration is empty
        outer += [ast.Assign(
            targets=[store_econtext(name)
                     for name in node.names],
            value=load("None"))
              ]

        # Compute inner body
        inner = self.visit(node.node)

        # After each iteration, decrease the index
        inner += template("index -= 1", index=index)

        # For items up to N - 1, emit repeat whitespace
        inner += template(
            "if INDEX > 0: append(WHITESPACE)",
            INDEX=index, WHITESPACE=ast.Str(s=node.whitespace)
            )

        # Main repeat loop
        outer += [ast.For(
            target=store("_item"),
            iter=load("_iterator"),
            body=assignment + inner,
            )]

        # Finally, clean up assignment if it's local
        if outer:
            outer += self._leave_assignment(names)

        self._scopes.pop()

        return outer

    def _get_translation_identifiers(self, name):
        assert self._translations
        prefix = id(self._translations[-1])
        stream = identifier("stream_%d" % prefix, name)
        append = identifier("append_%d" % prefix, name)
        return stream, append

    def _enter_assignment(self, names):
        for name in names:
            for stmt in template(
                "BACKUP = get(KEY, _marker)",
                BACKUP=identifier("backup_%s" % name, id(names)),
                KEY=ast.Str(s=fast_string(name)),
                ):
                yield stmt

    def _leave_assignment(self, names):
        for name in names:
            for stmt in template(
                "if BACKUP is _marker: del econtext[KEY]\n"
                "else:                 econtext[KEY] = BACKUP",
                BACKUP=identifier("backup_%s" % name, id(names)),
                KEY=ast.Str(s=fast_string(name)),
                ):
                yield stmt
